using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Windows.Forms;

namespace SmartLauncher
{
	static class SyncEngine
	{
		static SortedList<string, SyncItem> files;
		static SortedList<string, SyncItem> backup;

		public static event OnProcessStartDelegate ProcessStart;
		public static event OnProcessRunDelegate ProcessRun;
		public static event OnProcessStopDelegate ProcessStop;
		static string backupDir;

		#region "Constructors"
		static SyncEngine()
		{
			files = new SortedList<string, SyncItem>();
			backup = new SortedList<string, SyncItem>();

		}
		#endregion

		#region "Procedures"
		// TODO: SyncEngine.Compare(): RC should represent the files to update
		public static int Compare(string startPath, bool keepUnchanged)
		{
			SyncItem file;
			int countNone = 0;
			int countCopy = 0;
			int countDelete = 0;

			// =====================================================================
			// === Pre-Process                                                   ===
			// =====================================================================
			Trace.WriteLine(string.Format("START  SyncEngine.Compare('{0}', {1})", startPath, keepUnchanged));

			// --- Initialize data -------------------------------------------------
			files.Clear();
			backup.Clear();

			// =====================================================================
			// === Process                                                       ===
			// =====================================================================
			// --- Verify the start path exists ------------------------------------

			// --- Local Repository ------------------------------------------------
			// Vrifier que le chemin de dpart se trouve dans le repository
			startPath = Path.GetDirectoryName(startPath).ToLower();
			if (startPath.Contains(ApplicationStartInfo.LocalRoot) == false)
				throw new ArgumentException(string.Format("The path '{0}' is outside of the local repository.", startPath));

			// Rcuprer les fichiers dans le repository local
			if (RepositoryEngine.Scan(true, false) == false)
			{
				files.Clear();
				return 0;
			}
			foreach (RepositoryItem item in RepositoryEngine.Files.Values)
				files.Add(item.Key, new SyncItem(item, true, SyncActions.Delete));

			// --- Remote Repository -----------------------------------------------
			if (RepositoryEngine.Scan(false, false) == false)
			{
				files.Clear();
				return 0;
			}
			foreach (RepositoryItem item in RepositoryEngine.Files.Values)
			{
				if (files.ContainsKey(item.Key) == false)
				{
					// le fichier est nouveau: il faut le crer
					file = new SyncItem(item, false, SyncActions.Copy);
					file.IsNew = true;
					files.Add(file.RemoteFile.Key, file);
				}
				else
				{
					// le fichier existe: mise  jour ou inchang
					file = files[item.Key];
					if (file.LocalFile.VersionDate == item.VersionDate)
						if (keepUnchanged == true)
							file.Action = SyncActions.None;
						else
							files.Remove(item.Key);
					else
						file.Action = SyncActions.Copy;
				}
			}

			// =====================================================================
			// === Post-Process                                                  ===
			// =====================================================================
			// --- Enlever tous les fichiers  la racine ---------------------------
			for (int idx = 0; idx < files.Count; )
			{
				file = files.Values[idx];
				if ((file.FolderName == null) || (file.FolderName.Length == 0))
					files.RemoveAt(idx);
				else
					idx++;
			}

			// === Compute the number of files to be updated =======================
			foreach (SyncItem item in files.Values)
			{
				switch (item.Action)
				{
					case SyncActions.None:
						countNone++;
						break;
					case SyncActions.Copy:
						countCopy++;
						break;
					case SyncActions.Delete:
						countDelete++;
						break;
					default:
						break;
				}
			}

			Trace.WriteLine(string.Format("FINISH SyncEngine.Compare('{0}', {1}) (Total: {2}, None= {3}, Copy={4}, Delete={5})", startPath, keepUnchanged, files.Count, countNone, countCopy, countDelete));
			return (countCopy + countDelete); // files.Count;
		}

		public static bool Synchronize(bool doBackup)
		{
			int countCopy = 0;
			int countDelete = 0;

			Trace.WriteLine(string.Format("START  SyncEngine.Synchronize({0})", doBackup));

			if (doBackup == true)
				BeginTransaction();

			if (ProcessStart != null)
				ProcessStart(null, new ProcessEventArgs("Starting Repository Synchronization", files.Count));

			// Synchroniser les fichier
			try
			{
				int idx = 0;
				foreach (SyncItem item in files.Values)
				{
					if (ProcessRun != null)
						ProcessRun(null, new ProcessEventArgs(string.Format("Synchronizing {0}", item.FileName), ++idx));

					if (item.Action == SyncActions.Copy)
					{
						Trace.WriteLine(string.Format("   CPY {0}", item.RemoteFile.AbsolutePath));
						if (Directory.Exists(Path.GetDirectoryName(item.LocalFile.AbsolutePath)) == false)
							Directory.CreateDirectory(Path.GetDirectoryName(item.LocalFile.AbsolutePath));
						File.Copy(item.RemoteFile.AbsolutePath, item.LocalFile.AbsolutePath, true);
						countCopy++;
					}
					else
						if (item.Action == SyncActions.Delete)
						{
							Trace.WriteLine(string.Format("   DEL {0}", item.LocalFile.AbsolutePath));
							File.Delete(item.LocalFile.AbsolutePath);
							countDelete++;
						}
				}
			}
			catch (Exception ex)
			{
				Trace.WriteLine(string.Format("*** ERROR *** SyncEngine.Synchronize({0}) (Copy={1}, Delete={1})", doBackup, countCopy, countDelete));
				Trace.WriteLine(string.Format("              {0}", ex.Message));

				if (doBackup == true)
					RollbackTransaction();

				//	MessageBox.Show("The update process did not succeed", Properties.Settings.Default.ShellHome, MessageBoxButtons.OK, MessageBoxIcon.Error);
				throw ex;
			}

			if (ProcessStop != null)
				ProcessStop(null, new ProcessEventArgs("Removing empty folders", 0));
			RemoveEmptyFolders(ApplicationStartInfo.LocalRoot);

			if (ProcessStop != null)
				ProcessStop(null, new ProcessEventArgs("Deleting the back-up", 0));

			if (doBackup == true)
				CommitTransaction();

			if (ProcessStop != null)
				ProcessStop(null, new ProcessEventArgs("Synchronization completed", 0));

			Trace.WriteLine(string.Format("FINISH SyncEngine.Synchronize({0}) (Copy={1}, Delete={1})", doBackup, countCopy, countDelete));

			return true;
		}

		static void BeginTransaction()
		{
			Trace.WriteLine(string.Format("START  SyncEngine.BeginTransaction"));

			if (ProcessStart != null)
				ProcessStart(null, new ProcessEventArgs("Starting local repository backup", files.Count));

			backupDir = Path.Combine(Path.GetTempPath(), "IAC_" + DateTime.Now.ToString("yyyyMMddhhmmss"));
			if (Directory.Exists(backupDir) == true)
				Directory.Delete(backupDir, true);
			Directory.CreateDirectory(backupDir);

			int idx = 0;
			foreach (SyncItem item in files.Values)
			{
				if (ProcessRun != null)
					ProcessRun(null, new ProcessEventArgs(string.Format("Back-up of {0}", item.FileName), ++idx));

				if ((item.Action != SyncActions.None) && (File.Exists(item.LocalFile.AbsolutePath) == true))
				{
					SyncItem si = new SyncItem(item.LocalFile, true, item.Action);
					backup.Add(si.Key, si);
					si.RemoteFile.RepositoryPath = backupDir;

					if (Directory.Exists(Path.GetDirectoryName(si.RemoteFile.AbsolutePath)) == false)
						Directory.CreateDirectory(Path.GetDirectoryName(si.RemoteFile.AbsolutePath));
					File.Copy(item.LocalFile.AbsolutePath, si.RemoteFile.AbsolutePath, true);
				}
			}
			if (ProcessStop != null)
				ProcessStop(null, new ProcessEventArgs("Backup completed", 0));

			Trace.WriteLine(string.Format("FINISH SyncEngine.BeginTransaction"));
		}

		static void CommitTransaction()
		{
			Trace.WriteLine(string.Format("START  SyncEngine.CommitTransaction"));
			if (Directory.Exists(backupDir) == true)
				Directory.Delete(backupDir, true);
			Trace.WriteLine(string.Format("FINISH SyncEngine.CommitTransaction"));
		}

		static void RollbackTransaction()
		{
			int idx = 0;

			Trace.WriteLine(string.Format("START  SyncEngine.RollbackTransaction"));

			// === Restaurer les fichiers qui existaient en local ==================
			if (ProcessStart != null)
				ProcessStart(null, new ProcessEventArgs("Restoring local repository (1/2)", backup.Count));
			foreach (SyncItem item in backup.Values)
			{
				if (ProcessRun != null)
					ProcessRun(null, new ProcessEventArgs(string.Format("Restoring {0}", item.FileName), ++idx));
				try
				{
					File.Copy(item.RemoteFile.AbsolutePath, item.LocalFile.AbsolutePath, true);
				}
				catch { }
			}


			// === Effacer les nouveaux fichiers qui n'existaient pas en local =====
			if (ProcessStart != null)
				ProcessStart(null, new ProcessEventArgs("Restoring local repository (2/2)", files.Count));
			idx = 0;
			foreach (SyncItem item in files.Values)
			{
				if (ProcessRun != null)
					ProcessRun(null, new ProcessEventArgs(string.Format("Restoring {0}", item.FileName), ++idx));
				try
				{
					if (item.IsNew == true)
						File.Delete(item.LocalFile.AbsolutePath);
				}
				catch { }
			}

			if (ProcessStop != null)
				ProcessStop(null, new ProcessEventArgs("Restoring local repository", 0));

			Trace.WriteLine(string.Format("FINISH SyncEngine.RollbackTransaction"));
		}

		public static void RemoveEmptyFolders(string rootFolder)
		{
			string[] items;

			items = Directory.GetDirectories(rootFolder);
			foreach (string item in items)
				RemoveEmptyFolders(item);

			try
			{
				if (Directory.GetFiles(rootFolder).Length == 0)
					Directory.Delete(rootFolder);
			}
			catch { }
		}

		#endregion

		#region "Properties"
		public static SortedList<string, SyncItem> Files
		{
			get { return files; }
		}
		#endregion
	}
}
